/* parallellife.c */

/*  Copyright 2005 Dauger Research, Inc. 
    
Our lawyers made us say this: 
DISCLAIMER: We provide the following on an "AS IS" basis. Use it at your own risk.  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

void fractionalsleep(double sleepRequest); 
void fractionalsleep(double sleepRequest)
	{/* like sleep(), except that it accepts fractions of a second */
	if (sleepRequest>0) {
		struct timespec request; 
		double intSleep; 
		request.tv_nsec=(1e9*modf(sleepRequest, &intSleep)); 
		request.tv_sec=(long)intSleep; 
		nanosleep(&request, NULL); 
		}
	}

#ifndef Byte
#define Byte 		unsigned char
#endif

/* Defines array structure and access */
typedef struct LifeArrayStruct {
	Byte *array; 
	long rowBytes; /* access with array[col+rowBytes*row] */
	long columnSize; /* <= rowBytes-2 */
	long rowSize; 
	Byte buffer[4]; /* allocation must be at least sizeof(Byte)*(rowSize+2)*(columnSize+2) */
	} LifeArrayStruct, *LifeArrayPtr, **LifeArrayHandle; 

#include <string.h> 

LifeArrayPtr NewLifeArray(long columnSize, long rowSize); 
LifeArrayPtr NewLifeArray(long columnSize, long rowSize)
	{LifeArrayPtr out=NULL; 
	if ((columnSize>0)&&(rowSize>0)) {
		long rowBytes=columnSize+2; 
		long allocation=sizeof(LifeArrayStruct)+rowBytes*(rowSize+2)*sizeof(Byte); 
		out=malloc(allocation); 
		if (out) {
			memset(out, 0, allocation); 
			out->rowBytes=rowBytes; 
			out->columnSize=columnSize; 
			out->rowSize=rowSize; 
			out->array=&out->buffer[1+out->rowBytes]; 
			}
		}
	return out; 
	}

void propagatelife(LifeArrayPtr out, LifeArrayPtr in); /* assumed to be equal in size */
void propagatelife(LifeArrayPtr out, LifeArrayPtr in)
	{/* Propogates life according to rules by J. Conway in 1970 */
	if (out&&in) {
		long row=in->rowSize; 
		while (row--) {
			long column=in->columnSize; 
			Byte *outRow=&out->array[row*out->rowBytes];
			Byte *inRow=&in->array[row*in->rowBytes]; 
			
			/* Hey you! Vectorize this! */
			while (column--) {
				long neighborCount=(inRow[column-1]?1:0)+(inRow[column+1]?1:0); 
				neighborCount+=(inRow[column+in->rowBytes-1]?1:0)+(inRow[column+in->rowBytes]?1:0)+(inRow[column+in->rowBytes+1]?1:0); 
				neighborCount+=(inRow[column-1-in->rowBytes]?1:0)+(inRow[column-in->rowBytes]?1:0)+(inRow[column+1-in->rowBytes]?1:0); 
				
				switch (neighborCount) {
					default: 
					case 0: case 1: 
						/* not enough */
					case 4: case 5: case 6: case 7: case 8: case 9: 
						/* too much */
						outRow[column]=0; /* die */
						break; 
					case 2: if (!(inRow[column])) {
						outRow[column]=0; /* stay dead */
						break; 
						}
					case 3: /* just right */
						if (!(outRow[column]=inRow[column]+1)) outRow[column]=0xff; /* live */
						break; 
					}
				
				}
			
			}
		
		}
	
	}

#pragma mark - Parallel code follows 

#include "mpi.h"
int ppinit(int *idproc, int *nproc); 
void ppexit(void); 

void maintainboundaryconditions(LifeArrayPtr in, int idproc, int nproc); 
void maintainboundaryconditions(LifeArrayPtr in, int idproc, int nproc)
	{/* Maintains periodic boundary conditions */
	if (in) {
		long row=in->rowSize; 
		/* copy ends of rows */
		while (row--) {
			Byte *rowP=&in->array[row*in->rowBytes]; 
			rowP[-1]=rowP[in->columnSize-1]; 
			rowP[in->columnSize]=rowP[0]; 
			}
		
		/* copy first and last rows */
		{long column=in->columnSize+2; 
		Byte *firstRow=&in->array[-1]; 
		Byte *lastRow=&in->array[(in->rowSize-1)*in->rowBytes-1]; 
		
		if (nproc) {int ierr; MPI_Status status; 
			MPI_Request leftReq, rightReq; 
			int rightIDProc=idproc+1; 
			int leftIDProc=idproc-1; 
			
			if (rightIDProc>=nproc) rightIDProc=0; 
			if (leftIDProc<0) leftIDProc=nproc-1; 
			
			/*
			ierr=MPI_Irecv(&firstRow[-in->rowBytes], sizeof(Byte)*column, MPI_BYTE, 
				leftIDProc, leftIDProc, MPI_COMM_WORLD, &leftReq); 
			ierr=MPI_Irecv(&lastRow[+in->rowBytes], sizeof(Byte)*column, MPI_BYTE, 
				rightIDProc, rightIDProc, MPI_COMM_WORLD, &rightReq); 
			ierr=MPI_Send(lastRow, sizeof(Byte)*column, MPI_BYTE, 
				rightIDProc, idproc, MPI_COMM_WORLD); 
			ierr=MPI_Send(firstRow, sizeof(Byte)*column, MPI_BYTE, 
				leftIDProc, idproc, MPI_COMM_WORLD); 
			ierr=MPI_Wait(&leftReq, &status); 
			ierr=MPI_Wait(&rightReq, &status); 
			/*/
			
			{long i; 
			for(i=2; i--; ) 
				if (i^idproc&1) {
					ierr=MPI_Irecv(&firstRow[-in->rowBytes], sizeof(Byte)*column, MPI_BYTE, 
						leftIDProc, leftIDProc, MPI_COMM_WORLD, &leftReq); 
					ierr=MPI_Send(firstRow, sizeof(Byte)*column, MPI_BYTE, 
						leftIDProc, idproc, MPI_COMM_WORLD); 
					ierr=MPI_Wait(&leftReq, &status); 
					}
				else {
					ierr=MPI_Irecv(&lastRow[+in->rowBytes], sizeof(Byte)*column, MPI_BYTE, 
						rightIDProc, rightIDProc, MPI_COMM_WORLD, &rightReq); 
					ierr=MPI_Send(lastRow, sizeof(Byte)*column, MPI_BYTE, 
						rightIDProc, idproc, MPI_COMM_WORLD); 
					ierr=MPI_Wait(&rightReq, &status); 
					}
			}
			/**/
			
			}
		else 
			while (column--) {
				firstRow[column-in->rowBytes]=lastRow[column]; 
				lastRow[column+in->rowBytes]=firstRow[column]; 
				}
		
		}
		
		}
	}

#pragma mark - most of the parallel code is above

long fprintresult(FILE *stream, LifeArrayPtr in); 
long fprintresult(FILE *stream, LifeArrayPtr in)
	{long out=0; 
	if (in) {/* fprintf the array */
		long column; 
		for (column=0; column<in->columnSize; column++) {
			long row; 
			for (row=0; row<in->rowSize; row++) {
				char c=in->array[in->rowBytes*row+column]; 
				switch (c) {
					case 0: c=' '; break; 
					case 2: case 3: case 4: case 5: 
					case 6: case 7: case 8: case 9: case 10: 
						out++; 
					case 1: 
						out++; 
						c+='0'-1; break; 
					default: c='*'; break; 
					}
				if (stream) fprintf(stream, "%c", c); 
				}
			if (stream) fprintf(stream, "\n"); 
			}
		
		}
	return out>>1; 
	}

void saveresult(char *filename, LifeArrayPtr in); 
void saveresult(char *filename, LifeArrayPtr in)
	{/* save the array into a file */
	if (filename) {
		FILE *fp=fopen(filename, "a+"); 
		if (fp) {
			fprintresult(fp, in); 
			fclose(fp); 
			}
		}
	}

long genArray[]={0x0247, 0x07d9, 0x09be, 0x0fb9, 0x17d9, 0x1f4e, 0x1fdc, 0x27df, 0x8957}; 

void addmaterial(LifeArrayPtr in); 
void addmaterial(LifeArrayPtr in)
	{/* when called add material to the life array */
	if (in) {
		long rowstart=(rand()%(in->rowSize-6)); 
		long colstart=(rand()%(in->columnSize-6)); 
		if (rowstart<0) rowstart=-rowstart; 
		if (colstart<0) colstart=-colstart; 
		rowstart++; 
		colstart++; 
		
		{
		long j=4, s=(rand()&3)?0:(rand()%(sizeof(genArray)/sizeof(long))); 
		if (s<0) s=0; 
		
		s=genArray[s]; 
		
		printf("adding material %x\n", s); 
#ifndef __MWERKS__
		sleep(1); 
#endif
		
		if (rand()&1) {/* flip */
			s=((s&0xf000L)>>12)|
				((s&0x0f00L)>>4)|
				((s&0x0f0L)<<4)|
				((s&0x0fL)<<12); 
			}
		
		if (rand()&1) {/* flip */ 
			s=((s&0x8888L)>>3)|
				((s&0x4444L)>>1)|
				((s&0x2222L)<<1)|
				((s&0x1111L)<<3); 
			}
		
		while (j--) {
			long i=4; 
			while (i--) {
				in->array[i+colstart+(rowstart+j)*in->rowBytes]|=(s&1)<<2; 
				s>>=1; 
				}
			}
		
		}
		
		}
	}

#define RowDimension		80
#define ColumnDimension		22

void performcalculation(); 
void performcalculation()
	{
	/* allocate */
	LifeArrayPtr arrayA, arrayB; 
	arrayA=NewLifeArray(ColumnDimension, RowDimension); 
	arrayB=NewLifeArray(ColumnDimension, RowDimension); 
	if (arrayA&&arrayB) {
		long frameCount=0; 
		long countdown=8; 
		int nproc=0, idproc=0, ierr; 
		
		/* set initial conditions */ 
		{long row=RowDimension; 
		while (row--) {
			Byte *rowArray=&arrayA->array[row*arrayA->rowBytes]; 
			long column=ColumnDimension; 
			while (column--) {
				rowArray[column]=rand()&1; 
				}
			}
		}
		fprintresult(stdout, arrayA); 
		
		ierr=ppinit(&idproc, &nproc); 
		if (ierr) {
			printf("MPI_Init has returned error %d.  Assuming single-processor... \n", ierr); 
			nproc=idproc=0; 
			}
			
		while (frameCount<5000) {
			LifeArrayPtr lastArray=frameCount&1?arrayB:arrayA; 
			LifeArrayPtr nextArray=frameCount&1?arrayA:arrayB; 
			long activecellcount; 
			
			maintainboundaryconditions(lastArray, idproc, nproc); 
			propagatelife(nextArray, lastArray); 
			
			system("clear"); 
			
			printf("\fFrame %d idproc=%d\n", frameCount, idproc); 
			activecellcount=fprintresult(stdout, nextArray); 
			
			fractionalsleep(0.125); 
			
			if (activecellcount*100<(ColumnDimension*RowDimension)) {
				if (!countdown) {
					addmaterial(nextArray); 
					countdown=16; 
					}
				else 
					countdown--; 
				}
			else 
				countdown=32; 
			
			frameCount++; 
			
			}
		
		if (nproc) 
			ppexit(); 
		
		}
	
	/* deallocate */
	if (arrayA) free(arrayA); 
	if (arrayB) free(arrayB); 
	
	}

int main(void)
	{
	{
	long throwoutcount=time(NULL)&0x3ffL; 
	printf("Initializing random number generator by throwing out %d random numbers...\n", throwoutcount);
	while (throwoutcount--) rand(); 
	}
	printf("Beginning calculation... (first random is %d)\n", rand()); 
	
	performcalculation(); 
	
	return 0;
	}
	

#ifdef __MWERKS__
/* only for Metrowerks CodeWarrior */
#include <SIOUX.h>
#endif 

int ppinit(int *idproc, int *nproc)
	{
/* this subroutine initializes parallel processing
   idproc = processor id
   nproc = number of real or virtual processors obtained */
   int ierr;
   *nproc=0; 

/* initialize the MPI execution environment */

   ierr = MPI_Init(NULL, NULL);
   if (!ierr) {

		/* determine the rank of the calling process in the communicator */
		ierr = MPI_Comm_rank(MPI_COMM_WORLD, idproc);

		/* determine the size of the group associated with a communicator */
		ierr = MPI_Comm_size(MPI_COMM_WORLD, nproc);

		}
	
#ifdef __MWERKS__
/* only for Metrowerks CodeWarrior */
		SIOUXSettings.asktosaveonclose=0; 
		SIOUXSettings.autocloseonquit=1;
#endif 
		
	return ierr; 
	}

void ppexit(void)
	{
/* this subroutine terminates parallel processing */
	int ierr;
	
/* waits until everybody gets here */
	ierr = MPI_Barrier(MPI_COMM_WORLD); 
	
/* terminate MPI execution environment */
	ierr = MPI_Finalize();
	
	}
